/**************************************************************************
ganalyse procvides analytical tools for exploring the output of gprocess

April 2022

**************************************************************************/
using namespace std;
#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>
#include <algorithm>
#include <csignal>
#include <iomanip>
#include <ctype.h>
#include <functional>

/******************** Global Variables **********************
bool 			UNORIENTED_SEARCH = false;
bool			EQUIVALENCE_CLASS_DETAIL = false;
bool			WRITE_EQUIVALENCE_CLASSES = false;
bool			sigusr1_request = false;
bool			termination_request = false;
int				max_num_crossings=0;
int				num_components=0;
int 			hash_threshold=100;
string			filepath;

bool     		DEVELOPMENT_MODE = false; // used 

*/
extern int shift_power;
extern ofstream 	output;
extern ofstream    debug;

#include <ganalyse.h>

int string_to_int (string s);
void write_code_list_element (ostream& outstream, code_list_element& e);


/* class: 

   if symmetry-type == none	   
     <-1>,<min-gauss-code-representation>,<class-identifier>:<symmetry-type>:<symmetry-class-length>:<symmetry-class>
   if symmetry-type != none	   
     <-1>,<min-gauss-code-representation>,<class-identifier>:<symmetry-type>:<symmetry_representative>:<symmetry-class-length>:<symmetry-class>
     
   where <symmetry-class> = <element_first>:...:<element_last>
*/
void parse_equivalence_class(vector<equivalence_class>& code_class, string next_line)
{
	equivalence_class c;
	char ch;

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "parse_equivalence_class: presented with next_line " << next_line << endl;
	
	next_line = next_line.substr(next_line.find(':')+1); // discard the -1:

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "parse_equivalence_class: after removal of class indicator next_line = " << next_line << endl;

	size_t pos = next_line.find(':');
	c.min_representation = next_line.substr(0,pos);
	next_line = next_line.substr(pos+1);

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "parse_equivalence_class: after removal of min_representation next_line = " << next_line << endl;


	pos = c.min_representation.find('/');
	for (size_t p = pos; p < c.min_representation.length(); p++)
	{
		if (c.min_representation[p] == '+' || c.min_representation[p] == '-')
			c.min_num_crossings++;
	}

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "parse_equivalence_class: min_representation " << c.min_representation << ", min_num_crossings " << c.min_num_crossings << endl;


	istringstream iss(next_line);

	iss >> c.class_identifier;
	iss >> ch;
	iss >> c.symmetry_type;	
	iss >> ch;	
	if (c.symmetry_type != equivalence_class::symmetry::none)
	{
		iss >> c.symmetry_representative;
		iss >> ch;
	}
	
	int list_length = -1;
	iss >> list_length;

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "parse_equivalence_class: symmetry_class list length = " << list_length << endl;

	
	for (int i=0; i< list_length; i++)
	{
		iss >> ch;
		int element;
		iss >> element;
		c.symmetry_class.push_back(element);
	}
	
if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "parse_equivalence_class: read equivalence class:" << endl;
	debug << "parse_equivalence_class:   min_representation " << c.min_representation << endl;
	debug << "parse_equivalence_class:   min_num_crossings " << c.min_num_crossings << endl;
	debug << "parse_equivalence_class:   root_index " << c.root_index << endl;
	debug << "parse_equivalence_class:   class_identifier " << c.class_identifier << endl;
	debug << "parse_equivalence_class:   symmetry_type " << c.symmetry_type << endl;
	debug << "parse_equivalence_class:   symmetry representative " << c.symmetry_representative << endl;
	debug << "parse_equivalence_class:   symmetry class ";
	list<int>::iterator lptr = c.symmetry_class.begin();
	while (lptr != c.symmetry_class.end())
	{
		debug << *lptr << ' ';
		lptr++;
	}
	debug << endl;
}
	
	code_class[c.class_identifier] = c;
}

/* read the equivalence classes from the input file, which has a comma delimited format.  

	class: <-1>,<min-gauss-code-representation>,<class-identifier>

	node:  <index>,<parent-index>,<gauss-code>,<class-identifier>,<num-list-targets>,{<list-target>}

	where: <parent-index> is set to -1 if the node is the root of a class-tree
	       {<class-identifiers>} is a comma delimited set of integers
	       {<list-target>} is a comma delimted set of list-targets.
	       
	Each <list-target> has the form:
	
	<index>,<num-R1-moves>,<num-R2-moves>,<num-R3-moves>,{<R1-move>},{<R2-move>},{<R3-move>}
	
	where: <index> is set to -1 if the list target location is zero
	       <R1-move> = list of integers
	       <R2-move> = list of integer pairs
	       <R3-move> = list of integer triples (<ingress over-arc-edge> <ingress under-arc-edge-1> <ingress over-arc-edge-2>)			
*/	
void parse_code_list_element(vector<equivalence_class>& code_class, vector<code_list_element>& code_element, string next_line)
{
	
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "parse_code_list_element: next_line = " << next_line << endl;

	code_list_element e;

	istringstream iss(next_line);
	string field;
	
	/* read the first five fields */
	for (int i=0; i<5; i++)
	{
		getline(iss,field,':');
		
		switch(i)
		{
			case 0: e.index = string_to_int(field); break;
			case 1: e.parent = string_to_int(field); break;
			case 2: e.code = field; break;
			case 3: e.class_identifier = string_to_int(field); break;
			case 4: e.num_targets = string_to_int(field); break;
		}
	}

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "parse_code_list_element: index = " << e.index << " parent = " << e.parent << " code = " << e.code << " class_identifier = " << e.class_identifier << " num_targets = " << e.num_targets << endl;

	
	if (e.parent == -1)
	{
		code_class[e.class_identifier].root_index = e.index;
		
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "parse_code_list_element: node is the root of an equivalence class" << endl;
	}
	
	if (e.num_targets > 0)
		e.target = vector<list_target>(e.num_targets);
	
	for (int i=0; i< e.num_targets; i++)
	{
		list_target t;
		getline(iss,field,':');
		t.location = string_to_int(field);
		
		int num_R1_moves;
		getline(iss,field,':');
		num_R1_moves = string_to_int(field);

		int num_R2_moves;
		getline(iss,field,':');
		num_R2_moves = string_to_int(field);
		
		int num_R3_moves;
		getline(iss,field,':');
		num_R3_moves = string_to_int(field);

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "parse_code_list_element: list_target location = " << t.location << " num_R1_moves = " << num_R1_moves << " num_R2_moves = " << num_R2_moves << " num_R3_moves = " << num_R3_moves;

		
		if (num_R1_moves > 0)
		{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << " R1 moves ";
	
			t.R1_move = vector<int>(num_R1_moves);
			for (int i=0; i< num_R1_moves; i++)
			{
				getline(iss,field,':');
				t.R1_move[i] = string_to_int(field);				
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << t.R1_move[i] << ' ';
			}
		}
		
		if (num_R2_moves > 0)
		{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << " R2 moves ";
			t.R2_move = matrix<int>(num_R2_moves,2);
			for (int i=0; i< num_R2_moves; i++)
			{
				getline(iss,field,':');
				t.R2_move[i][0] = string_to_int(field);				
				getline(iss,field,':');
				t.R2_move[i][1] = string_to_int(field);				
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << t.R2_move[i][0] << ':' << t.R2_move[i][1] << ' ';
			}
		}
		
		if (num_R3_moves > 0)
		{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << " R3 moves ";
			t.R3_move = matrix<int>(num_R3_moves,3);
			for (int i=0; i< num_R3_moves; i++)
			{
				getline(iss,field,':');
				t.R3_move[i][0] = string_to_int(field);				
				getline(iss,field,':');
				t.R3_move[i][1] = string_to_int(field);				
				getline(iss,field,':');
				t.R3_move[i][2] = string_to_int(field);				
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << t.R3_move[i][0] << ':' << t.R3_move[i][1] << ':' << t.R3_move[i][2] << ' ';
			}
		}
		
		e.target[i] = t;

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << endl;
	}
	
	code_element[e.index] = e;		
}

int string_to_int (string s)
{
	istringstream iss(s);
	int i;
	iss >> i;
	return i;
}

void print_equivalence_class(equivalence_class& e, ofstream& os, string prefix)
{
	os << prefix << "min_representation: " << e.min_representation << endl;
	os << prefix << "min_num_crossings: " << e.min_num_crossings << endl;
	os << prefix << "root_index: " << e.root_index << endl;
	os << prefix << "class_identifier: " << e.class_identifier << endl;
	os << prefix << "symmetry_type: " << e.symmetry_type << endl;
	os << prefix << "symmetry_representative: " << e.symmetry_representative << endl;
	os << prefix << "symmetry_class: ";

	if (e.symmetry_class.size() != 0)
	{
		list<int>::iterator lptr = e.symmetry_class.begin();
		while (lptr != e.symmetry_class.end())
		{
			os << ':' << *lptr;
			lptr++;
		}
		os << endl;
	}
	else
	{
		os << "empty" << endl;
	}
}

void write_equivalence_classes(vector<equivalence_class>& code_class, vector<code_list_element>& code_element, ofstream& os)
{
	vector<bool> visited(code_element.size());
	
	for(unsigned int i=0; i< code_class.size(); i++)
	{
		visited[code_class[i].root_index] = true;
		
		os << -1 << ':' << code_class[i].min_representation << ':' << code_class[i].class_identifier; 

		if (code_class[i].min_representation != "E")
		{	
			os << ':' << code_class[i].symmetry_type << ':';
			
			if (code_class[i].symmetry_type != equivalence_class::symmetry::none)
				os << code_class[i].symmetry_representative << ':';
				
			os << code_class[i].symmetry_class.size();
			
			if (code_class[i].symmetry_class.size() != 0)
			{
				list<int>::iterator lptr = code_class[i].symmetry_class.begin();
				while (lptr != code_class[i].symmetry_class.end())
				{
					os << ':' << *lptr;
					lptr++;
				}
			}
		}
		
		os << endl;

		write_code_list_element(os,code_element[code_class[i].root_index]);

		if (code_element[code_class[i].root_index].num_targets != 0)
		{			
			
			list<int> class_target;
			
			for (int j=0; j< code_element[code_class[i].root_index].num_targets; j++)
			{
				if (code_element[code_class[i].root_index].target[j].location != -1)
				{
					class_target.push_back(code_element[code_class[i].root_index].target[j].location);					
if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "write_equivalence_classes: class_target.push_back  target " << j << " from node " << code_class[i].root_index 
	      << ", the root of class " << i << ": " << code_element[code_class[i].root_index].target[j].location << endl;
}
				}
			}
				
			while (class_target.size() > 0)
			{
				int next_node = *(class_target.begin());

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "write_equivalence_classes: next_node " << next_node << endl;

				/* a location may have been pushed onto the class-target list more than once before we had a chance to process it */
				if (visited[next_node])
				{
					class_target.pop_front();
					continue;
				}

				visited[next_node] = true;
				
				write_code_list_element(output,code_element[next_node]);

				class_target.pop_front();
				
				for (int j=0; j< code_element[next_node].num_targets; j++)
				{
					if (code_element[next_node].target[j].location != -1 && !visited[code_element[next_node].target[j].location])
					{
						class_target.push_back(code_element[next_node].target[j].location);			
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "write_equivalence_classes: class_target.push_back target " << j << " from node " << next_node << ": " << code_element[next_node].target[j].location << endl;
					}
				}				
			}				
		}	
	}
}

void write_code_list_element (ostream& outstream, code_list_element& e)   
{
	outstream << e.index << ':' << e.parent << ':' << e.code << ':'<< e.class_identifier << ':' << e.num_targets;
	
	if (e.num_targets != 0)
		outstream << ':';
		
	for (int i=0; i< e.num_targets; i++)
	{
		write_list_target(outstream,e.target[i]);
		if (i < e.num_targets -1)
			outstream << ':';
	}
	outstream << endl;
}

void write_list_target(ostream& outstream, list_target& t) 
{
	outstream << t.location << ':' << t.R1_move.size() << ':' << t.R2_move.numrows() << ':' << t.R3_move.numrows();
	
	if (t.R1_move.size())
		outstream << ':';

	for (unsigned int i=0; i< t.R1_move.size(); i++)
	{
		outstream << t.R1_move[i];
		if (i < t.R1_move.size()-1)
			outstream << ':';
	}

	if (t.R2_move.numrows())
		outstream << ':';
		
	for (unsigned int i=0; i< t.R2_move.numrows(); i++)
	{
		outstream << t.R2_move[i][0] << ':' << t.R2_move[i][1];
		if (i < t.R2_move.numrows()-1)
			outstream << ':';
	}

	if (t.R3_move.numrows())
		outstream << ':';

	for (unsigned int i=0; i< t.R3_move.numrows(); i++)
	{
		outstream << t.R3_move[i][0] << ':' << t.R3_move[i][1] << ':' << t.R3_move[i][2];
		if (i < t.R3_move.numrows()-1)
			outstream << ':';
	}
}

void print_list_target (ostream& outstream, list_target& t, string prefix)
{
	outstream << prefix << "location = " << t.location << ", num R1 moves = " << t.R1_move.size() << ", num R2 moves = " << t.R2_move.numrows() << ", num R3 moves = " << t.R3_move.numrows() << endl;
	outstream << prefix << "R1 moves: ";
	for (unsigned int i=0; i< t.R1_move.size(); i++)
	{
		outstream << t.R1_move[i];
		if (i < t.R1_move.size()-1)
			outstream << ' ';
	}
	outstream << endl;
	outstream << prefix << "R2 moves: ";
	for (unsigned int i=0; i< t.R2_move.numrows(); i++)
	{
		outstream << t.R2_move[i][0] << ',' << t.R2_move[i][1];
		if (i < t.R2_move.numrows()-1)
			outstream << ' ';
	}
	outstream << endl;
	outstream << prefix << "R3 moves: ";
	for (unsigned int i=0; i< t.R3_move.numrows(); i++)
	{
		outstream << t.R3_move[i][0] << ',' << t.R3_move[i][1] << ',' << t.R3_move[i][2] << ',' << t.R3_move[i][3] << ',' << t.R3_move[i][4] << ',' << t.R3_move[i][5];
		if (i < t.R3_move.numrows()-1)
			outstream << ' ';
	}
	outstream << endl;
}

string minimum_class_representative(vector<list<int> >& code_hash_list, vector<equivalence_class>& code_class, vector<code_list_element>& code_element, string s)
{
	bool found = false;
	string min_representation;
	
	size_t hash_val = hash<string>()(s);
	hash_val >>= 64-shift_power; 
	
	/* initialize cptr to the start of the list of indices into code_element corresponding to those codes having the same hash value */
	list<int>::iterator cptr = code_hash_list[hash_val].begin();
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "minimum_class_representative: hash_val for " << s <<  " = " << hash_val << endl;
	
	/* now we only have to search a few specific indices in code_element when looking for s */
	while (cptr != code_hash_list[hash_val].end())
	{
		if (code_element[*cptr].code == s)
		{
			found = true;
			
			/* code_element[*cptr] is the code_list_element for s */
			min_representation = code_class[code_element[*cptr].class_identifier].min_representation;
			break;
		}
		cptr++;
	}
	
	if (!found)
	{
		cout << "Error! minimum_class_representative failed to find " << s << endl;
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "minimum_class_representative: Error! minimum_class_representative failed to find " << s << endl;
	
		exit(0);
	}
	else
	{
		return min_representation;
	}
}

/* code_element_index returns the index into code_element of the code_list_element corresponding to the Gauss code s */
int code_element_index(vector<list<int> >& code_hash_list, vector<code_list_element>& code_element, string s)
{

	bool found = false;
	int index = -1;
	
	size_t hash_val = hash<string>()(s);
	hash_val >>= 64-shift_power; 
	
	/* initialize cptr to the start of the list of indices into code_element corresponding to those codes having the same hash value */
	list<int>::iterator cptr = code_hash_list[hash_val].begin();
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "code_element_index: hash_val for " << s <<  " = " << hash_val << endl;
	
	/* now we only have to search a few specific indices in code_element when looking for s */
	while (cptr != code_hash_list[hash_val].end())
	{
		if (code_element[*cptr].code == s)
		{
			found = true;
			
			/* code_element[*cptr] is the code_list_element for s */
			index = *cptr;
			break;
		}
		cptr++;
	}
	
	if (!found)
	{
		cout << "Error! code_element_index failed to find " << s << endl;
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "code_element_index: Error! code_element_index failed to find " << s << endl;
	
		exit(0);
	}
	else
	{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "code_element_index: code_element_index = " << index << endl;
	
		return index;
	}
}

vector<int> count_equivalence_class_members(vector<code_list_element>& code_element, int num_code_list_elements, int num_equivalence_classes)
{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "count_equivalence_class_members: code_element.size() = " << code_element.size() << " num_code_list_elements = " << num_code_list_elements << " num_equivalence_classes = " << num_equivalence_classes << endl;

	vector<int> class_element_count(num_equivalence_classes);
	vector<bool> visited(num_code_list_elements);
	
	for (int i=0; i< num_code_list_elements; i++)
	{
		if (!visited[code_element[i].index])
		{
			class_element_count[code_element[i].class_identifier]++;
			visited[code_element[i].index] = true;
		}
	}	
	
	return class_element_count;
}

/* tree_node_index searches the first node_count entries of tree_node looking for a Reidemeister_tree_node with index equal to the index supplied as a parameter.
   if the search is succesful the index into tree_node of the corresponding Reidemeister_tree_node is returned, otherwise the function returns -1.
   
   Note that his search is correct even if the index parameter is already -1, that is, if we are searching for the tree node corresponding to the unlink.
*/
int tree_node_index(vector<Reidemeister_tree_node>& tree_node,int node_count, int index)
{
	for (int i=0; i< node_count; i++)
	{
		if (tree_node[i].index == index)
			return i;
	}
	return -1;
}

/* preferred_parent checks whether tree_node offers neighbour_tree_node a better path to the root via neighbour_target 

	we prefer this_node if it provides:

		fewer hops to the root
		fewer Reidemeister II moves
		fewer Reidemeister I and III moves

*/
bool preferred_parent(Reidemeister_tree_node& tree_node, list_target& neighbour_target, Reidemeister_tree_node& neighbour_tree_node)
{
	bool preferred = false;
	
	if (neighbour_tree_node.root_hop_count > tree_node.root_hop_count +1)
	{
		preferred = true;
	}
	else if (neighbour_tree_node.root_hop_count == tree_node.root_hop_count +1)
	{
		if (neighbour_tree_node.num_R2_moves > tree_node.num_R2_moves + static_cast<int>(neighbour_target.R2_move.size()))
		{
			/* this_node offers a path involving fewer Reidemister II moves */
			preferred = true;
		}
		else if (neighbour_tree_node.num_R2_moves == tree_node.num_R2_moves + static_cast<int>(neighbour_target.R2_move.size()))
		{
			if (neighbour_tree_node.num_R1_moves + neighbour_tree_node.num_R3_moves > tree_node.num_R1_moves + static_cast<int>(neighbour_target.R1_move.size()) + tree_node.num_R3_moves + static_cast<int>(neighbour_target.R3_move.size()))
			{
				preferred = true;
			}
		} 
		
	}							
	
	return preferred;
}


void print_Reidemeister_moves(ostream& outstream, list_target& target)
{
	for (size_t i=0; i< target.R3_move.numrows(); i++)
	{
		outstream << "    Reidemeister III move involving over-arc crossings " << target.R3_move[i][0] << " and " << target.R3_move[i][1];
		outstream << " and crossing " << target.R3_move[i][2] << endl;
	}
	
	for (size_t i=0; i< target.R2_move.numrows(); i++)
	{
		outstream << "    Reidemeister II move ";
		for (int j=0; j< 2; j++)
			outstream << target.R2_move[i][j] << ' ';
		outstream << endl;
	}

	for (size_t i=0; i< target.R1_move.size(); i++)
	{
		outstream << "    Reidemeister I move ";
		outstream << target.R1_move[i] << endl;
	}
}

void print_Reidemeister_tree_node (ostream& outstream, Reidemeister_tree_node& n, string prefix)
{
	outstream << prefix << "index = " << n.index << endl;
	outstream << prefix << "parent = " << n.parent << endl;
	outstream << prefix << "parent_tree_node_offset = " << n.parent_tree_node_offset << endl;
	outstream << prefix << "root_hop_count = " << n.root_hop_count << endl;
	outstream << prefix << "root_list_target = " << n.root_list_target << ", parent_list_target = " << n.parent_list_target << endl;
	outstream << prefix << "num_R1_moves = " << n.num_R1_moves << ", num_R2_moves = " << n.num_R2_moves << ", num_R3_moves = " << n.num_R3_moves << endl;
}
