/************************************************************************
                  Support functions for glist and gprocess

ostream& operator << (ostream& outstream, list_target& t)
void write_list_target (ostream& outstream, list_target& t)
void print_code_list_element (ostream& outstream, code_list_element& e, string prefix)   
void write_code_list_element (ostream& outstream, code_list_element& e,int class_index, bool OU_format)   
bool next_label(vector<int>& array, int& num_classifying_crossings)
int Reidemeister_I_present(gauss_orientation_data& gauss_data)
bool classical_Reidemeister_II_present(gauss_orientation_data& gauss_data, pair<int,int>& bigon)
void locate_code(list_target& target, vector<list<code_list_element> >& code_list, bool hashed, string preferred_moved_code, int num_crossings)
void remove_Reidemeister_I(generic_code_data& code_data, list_target& target)
code_list_element* find_next_root(vector<vector<list<code_list_element> >*>& code_list_pointer)
bool equivalence_class::change_to_representation(string& s)
void reverse_root_path(code_list_element* tree_node, code_list_element* new_parent)
void equivalence_class_location_error(int class_identifier, string message)
generic_code_data remove_Reidemeister_II(gauss_orientation_data gauss_data, pair<int,int>& bigon, list_target& t)
bool opgc_less_than(string s1, string s2)


**************************************************************************/
using namespace std;
#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>
#include <algorithm>
#include <iomanip>

/********************* External variables ***********************/
extern bool			UNORIENTED_SEARCH;
extern bool			WRITE_ELEMENT_POINTERS;
extern ofstream     debug;
extern ofstream     output;
extern int			max_num_crossings;
extern int			num_components;
extern int 			hash_threshold;
extern int 			shift_power;
extern string		filepath;
extern bool     	DEVELOPMENT_MODE;



#include <glist.h>

#define ONE_ROW_AND_COLUMN		1

/********************* Function prototypes ***********************/
bool check_gauss_data(generic_code_data& lcode_data, generic_code_data& rcode_data);
bool equivalent_braids (string b1, string b2);


namespace util {
	string itos (long n);
}




ostream& operator << (ostream& outstream, list_target& t)
{
	outstream << "location = " << t.location;
	return outstream;
}

void write_list_target (ostream& outstream, list_target& t)
{
	if (t.location == 0)
		outstream << -1; 
	else
		outstream << t.location->index;
		
	outstream << ':' << t.R1_move.size() << ':' << t.R2_move.size() << ':' << t.R3_move.size();
	
	if (t.R1_move.size())
		outstream << ':';

	list<int>::iterator R1_ptr = t.R1_move.begin();
	for (unsigned int i=0; i< t.R1_move.size(); i++)
	{
		outstream << *R1_ptr;
		if (i < t.R1_move.size()-1)
			outstream << ':';
		R1_ptr++;
	}

	if (t.R2_move.size())
		outstream << ':';
		
	list<pair<int,int> >::iterator R2_ptr = t.R2_move.begin();
	for (unsigned int i=0; i< t.R2_move.size(); i++)
	{
		outstream << R2_ptr->first << ':' << R2_ptr->second;
		if (i < t.R2_move.size()-1)
			outstream << ':';
		R2_ptr++;
	}

	if (t.R3_move.size())
		outstream << ':';

	list<vector<int> >::iterator R3_ptr = t.R3_move.begin();
	for (unsigned int i=0; i< t.R3_move.size(); i++)
	{
		outstream << (*R3_ptr)[3] << ':' << (*R3_ptr)[4] << ':' << (*R3_ptr)[5];
		if (i < t.R3_move.size()-1)
			outstream << ':';
		R3_ptr++;
	}


}

void print_code_list_element (ostream& outstream, code_list_element& e, string prefix)   
{
	outstream << prefix << "code " << e.code << ", index = " << e.index << " class identifier = " << e.member_equivalence_class << " parent = " << e.parent << endl;
	outstream << prefix << "num_targets = " << e.num_targets << endl;
	for (int i=0; i< e.num_targets; i++)
		outstream << prefix << i+1 << ". " << e.target[i] << endl;
}

void write_code_list_element (ostream& outstream, code_list_element& e,int class_index)   
{
	outstream << e.index << ':';
	if (e.parent == 0)
		outstream << -1;
	else
		outstream << e.parent->index;
		
	outstream << ':';
	outstream << e.code;
	
	outstream << ':';
	if (class_index == -1)
		outstream << e.member_equivalence_class;
	else
		outstream << class_index;
		
	outstream << ':' << e.num_targets;
	
	if (e.num_targets != 0)
		outstream << ':';
		
	for (int i=0; i< e.num_targets; i++)
	{
		write_list_target(outstream,e.target[i]);
		if (i < e.num_targets -1)
			outstream << ':';
	}
	outstream << endl;
}

/* next_label evaluates the next set of labels by considering
   generic_code_data::VIRTUAL=0 generic_code_data::POSITIVE=1 and generic_code_data::NEGATIVE=2 performing the equivalent
   of tertiary addition

   When we're not ignoring virtual crossings, we must ensure that there is at least one virtual before we return
*/
bool next_label(vector<int>& array, int& num_classifying_crossings)
{
	bool found;
	bool look_again;

if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "next_label: given labels " ;
	for (unsigned int i=0; i< array.size(); i++)
		debug << array[i] << " ";
	debug << endl;
}

	do
	{
		found = false;
		
		for (int i=array.size()-1; i>=0; i--)
		{
			{
				if (array[i] == generic_code_data::VIRTUAL)
				{
					continue;
				}
				else if (array[i] == generic_code_data::POSITIVE)
				{
					array[i] = generic_code_data::NEGATIVE;
					found = true;
					break;
				}
				else
				{
					array[i] = generic_code_data::POSITIVE;
				}
			}
		}
		
		look_again = false;
		
	} while (look_again);
	
	
if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "next_label: returning labels " ;
	for (unsigned int i=0; i< array.size(); i++)
		debug << array[i] << " ";
	debug << endl;
}

	return found;
}

int Reidemeister_I_present(gauss_orientation_data& gauss_data)
{
	int num_terms = gauss_data.num_terms;
	matrix<int>& orientation_matrix = gauss_data.orientation_matrix;
	vector<int> start_of_component = gauss_data.start_of_component;
	vector<int> num_terms_in_component = gauss_data.num_terms_in_component;

	int ith_term_component=0;
	for (int i=0; i< num_terms; i++)
	{

		if (i == start_of_component[ith_term_component]+num_terms_in_component[ith_term_component])
			ith_term_component++;
			
		int succeeding_term;

		if (i == start_of_component[ith_term_component]+num_terms_in_component[ith_term_component]-1) // last term in component
			succeeding_term = start_of_component[ith_term_component];
		else
			succeeding_term = i+1;		
		
		int succeeding_term_component = 0;
		
		for (int j=1; j<gauss_data.num_components; j++)
		{
			if (succeeding_term >= start_of_component[j])
				succeeding_term_component++;
			else
				break;
		}
		
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "Reidemeister_I_present: term " << i << ", component " << ith_term_component << ", succeeding term " << succeeding_term << ", component " << succeeding_term_component << endl;

		if (ith_term_component == succeeding_term_component && succeeding_term != i && orientation_matrix[1][i] == orientation_matrix[1][succeeding_term])
		{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "Reidemeister_I_present: found Reidemeister I move in gauss_data term " << i << endl;
			return orientation_matrix[1][i];
		}
	}

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "Reidemeister_I_present: no Reidemeister I move found" << endl;
	
	return 0;
}

bool classical_Reidemeister_II_present(gauss_orientation_data& gauss_data, pair<int,int>& bigon)
{

if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "classical_Reidemeister_II_present: presented with gauss_data:" ;
	print_gauss_data(gauss_data, debug, "classical_Reidemeister_II_present:   ");
}

	vector<int>& classical_gauss_data = gauss_data.classical_gauss_data;
	vector<int>& classical_crossing_sign = gauss_data.classical_crossing_sign;

	int num_terms = gauss_data.num_terms;
	matrix<int>& orientation_matrix = gauss_data.orientation_matrix;
	vector<int> start_of_component = gauss_data.start_of_component;
	vector<int> num_terms_in_component = gauss_data.num_terms_in_component;

	int start_gauss_crossing;
	int rendezvous_gauss_crossing;
	int rendezvous_start_term;  // used for debugging

	int first_component = 0;

	bool Reidemeister_II_found = false;
	
	for (int i=0; i< num_terms-1; i++)
	{
		if (i == start_of_component[first_component]+num_terms_in_component[first_component])
			first_component++;
		
		int successor = (i+1-start_of_component[first_component])%num_terms_in_component[first_component] + start_of_component[first_component];
								
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "term " << i << ", first_component " << first_component << ", succeeding term " << successor << endl;
										
		if (classical_gauss_data[i] * classical_gauss_data[successor] > 0) // over-over or under-under sequence
		{
			start_gauss_crossing = orientation_matrix[1][i]-1;
			rendezvous_gauss_crossing = orientation_matrix[1][successor]-1;
													
if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "potential RII sequence found at term " << i << endl;
	debug << "  start_gauss_crossing " << start_gauss_crossing+1 << endl; 
	debug << "  rendezvous_gauss_crossing " << rendezvous_gauss_crossing+1 << endl; 
}							
			if (classical_crossing_sign[start_gauss_crossing]*classical_crossing_sign[rendezvous_gauss_crossing] > 0) // need opposite signs at each crossing
				continue;

			int second_component = first_component;
			
			/* if we have found a first pair of a potential Reidemeister II detour at the end and beginning of a
			   component, we start looking for the second pair at the first term of the next component.  Otherwise
			   we start looking at the term after the (consecutive) pair of terms.
			   
			*/
			int second_start_term = (successor == i+1? i+2: i+1);  // if i+1 or i+2 == num_terms the following loop doesn't execute
			for (int j=second_start_term; j<num_terms; j++)
			{
				if (j == start_of_component[second_component]+num_terms_in_component[second_component])
					second_component++;

				successor = (j+1-start_of_component[second_component])%num_terms_in_component[second_component] + start_of_component[second_component];
								
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "  term " << j << ", second_component " << second_component << ", succeeding term " << successor << endl;
													
				/* In orientation data such as R1 L2 L3 L1 R3 R2, it is possible that the successor of j(=5) is the starting point i(=0) but there is
				   no Reidemeister II detour present; that is, i, i-successor, j, and j-successor must be distinct.

				   is              i         j js					   
				   R1 R2 L2 R3 R4 L5, L3 L4 R5 L1

				   We need only check that j-successor is distinct from i, since i, i-successor and j will be distinct by definition 
				*/
				
				if (successor != i)
				{
					if (orientation_matrix[1][j]-1 == start_gauss_crossing)
					{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "    term " << j << " involves start_gauss_crossing" << endl;
											
						if (orientation_matrix[1][successor]-1 == rendezvous_gauss_crossing)
						{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "    succeeding term " << successor << " involves rendezvous_gauss_crossing" << endl;
								
							rendezvous_start_term = j;
								
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "Reidemeister_II configuration found, with rendezvous starting at term " << j << ", moving forwards from start_peer " << endl;

							Reidemeister_II_found = true;
							break;
						}
						else
						{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "    succeeding term " << successor << " does not involve rendezvous_gauss_crossing" << endl;
						}
					}
					else if (orientation_matrix[1][j]-1 == rendezvous_gauss_crossing)
					{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "    term " << j << " involves rendezvous_gauss_crossing" << endl;
									
						if (orientation_matrix[1][successor]-1 == start_gauss_crossing)
						{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "    succeeding term " << successor << " involves start_gauss_crossing" << endl;
								
							rendezvous_start_term = j;
								
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "Reidemeister_II configuration found, with rendezvous starting at term " << j << ", moving backwards from start_peer " << endl;
								
							Reidemeister_II_found = true;
							break;
						}
						else
						{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "    succeeding term " << successor << " does not involve start_gauss_crossing" << endl;
						}
					}
					else
					{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "    term " << j << " does not involve start_gauss_crossing or rendezvous_gauss_crossing" << endl;
					}
				}
			}
						
			if (Reidemeister_II_found)
			{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "Reidemeister_II found, with rendezvous starting at term " << rendezvous_start_term << endl;
			}
			else
			{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "no corresponding pair of terms found, not a Gauss_Reidemeister_II" << endl;
			}			
		}
		
		if (Reidemeister_II_found)
		{
			bigon.first = start_gauss_crossing+1;
			bigon.second = rendezvous_gauss_crossing+1;
			break;			
		}
	}

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "Reidemeister_II found  = " << Reidemeister_II_found << endl;
	
	return Reidemeister_II_found;

}
							

void locate_code(list_target& target, vector<list<code_list_element> >& code_list, bool hashed, string preferred_moved_code, int num_crossings)
{	
	if (hashed)
	{
		size_t hash_val = hash<string>()(preferred_moved_code);
		hash_val >>= 64-shift_power;
							
		list<code_list_element>& hash_list = code_list[hash_val];
		list<code_list_element>::iterator sptr = find_if(hash_list.begin(),hash_list.end(),code_equals(preferred_moved_code));
	
		if (sptr != hash_list.end())
		{							
			int offset = distance(hash_list.begin(),sptr);
							
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "locate_code: found preferred moved code at code_list index " << hash_val << ", offset " << offset << " of the list" << endl;
	
			target.location = &*sptr;	
		}
		else
		{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "Error!  locate_code failed to find over preferred code " << preferred_moved_code << " at hash_value " << hash_val << "in list for " << num_crossings << " crossings" << endl;

			cout << "Error!  locate_code failed to find over preferred code " << preferred_moved_code << " in list for " << num_crossings << " crossings" << endl;
			exit(0);
		}
	}
	else
	{

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "locate_code: preferred_moved_code = '" << preferred_moved_code << "'" << endl;
		
		bool found = false;
		for (size_t k=0; k < code_list.size(); k++)
		{
			list<code_list_element>::iterator sptr =  code_list[k].begin();
			if (sptr->code == preferred_moved_code)
			{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "found preferred moved code at location " << k << ", offset " << 0 << " of the list" << endl;
	
				target.location = &*sptr;	
				found = true;
				break;
			}
			else
			{
//if (debug_control::DEBUG >= debug_control::SUMMARY)
//	debug << "locate_code: code_list code '" << sptr->code  << "' differs" << endl;
			}
		}
		
		if (!found)
		{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "Error!  locate_code failed to find over preferred code " << preferred_moved_code << " in non-hashed list for " << num_crossings << " crossings" << endl;
			cout << "Error!  locate_code failed to find over preferred code " << preferred_moved_code << " in list for " << num_crossings << " crossings" << endl;
			exit(0);
		}								
	}
}



void remove_Reidemeister_I(generic_code_data& code_data, list_target& target)
{	
	gauss_orientation_data gauss_data(code_data);		
	int R1_gauss_crossing = Reidemeister_I_present(gauss_data);
		
	if (R1_gauss_crossing !=0)
	{							
		int R1_count = 0; // for debugging only 
		
		do
		{
			R1_count++;
			target.R1_move.push_back(R1_gauss_crossing);
			
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "remove_Reidemeister_I: moved code contains a simplifying Reidemeister I loop involving Gauss crosing " << R1_gauss_crossing << endl;
	
			ostringstream oss;
			
			for (int k=0; k<gauss_data.num_components; k++)
			{
				int term_count = 0;
				for (int l=gauss_data.start_of_component[k]; l<gauss_data.start_of_component[k]+gauss_data.num_terms_in_component[k]; l++)
				{
					bool keep_term = false;
					if (abs(gauss_data.classical_gauss_data[l]) < R1_gauss_crossing)
					{
						keep_term = true;
						oss << gauss_data.classical_gauss_data[l];
						term_count++;
					}
					else if (abs(gauss_data.classical_gauss_data[l]) > R1_gauss_crossing)
					{
						keep_term = true;
						if (gauss_data.classical_gauss_data[l] > 0)
							oss << gauss_data.classical_gauss_data[l]-1;									
						else
							oss << gauss_data.classical_gauss_data[l]+1;									
						term_count++;
					}
					
					if (keep_term && (l < gauss_data.start_of_component[k]+gauss_data.num_terms_in_component[k]-1))
						oss << ' ';
				}

				if (term_count > 0 && k<gauss_data.num_components-1)
					oss << ',';
			}
			
			oss << '/';
			
			for (int k=0; k<gauss_data.num_terms/2; k++)
			{
				if (k != R1_gauss_crossing-1)
				{
					if (gauss_data.classical_crossing_sign[k] > 0)
						oss << "+ ";
					else									
						oss << "- ";
				}
			}
											
			string reduced_code = oss.str();
			size_t pos = reduced_code.find(",/");
			if (pos != string::npos)
				reduced_code.erase(pos,1);
														
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "Reidemeister I reduced Gauss code " << reduced_code << endl;

			read_gauss_code(code_data, reduced_code);

			if (code_data.num_crossings > 0)
			{
				gauss_data = gauss_orientation_data(code_data);
				R1_gauss_crossing = Reidemeister_I_present(gauss_data);
			}
			else
			{
				break;
			}
		
		} while (R1_gauss_crossing !=0 );

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "remove_Reidemeister_I: total number of Reidemeister I loops = " << R1_count << endl;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "remove_Reidemeister_I: resultant code_data: ";
	write_gauss_code(debug,code_data,false);
	debug << endl;
}						

	}
}


code_list_element* find_next_root(vector<vector<list<code_list_element> >*>& code_list_pointer)
{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "find_next_root: code_list_pointer.size() = " << code_list_pointer.size() << endl;
	
	for (unsigned int num_crossings = 2; num_crossings < code_list_pointer.size(); num_crossings++)
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "find_next_root: num_crossings = " << num_crossings << " code_list_pointer[num_crossings] = " << code_list_pointer[num_crossings] << endl;

		if (code_list_pointer[num_crossings] != 0)
		{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "find_next_root: code_list_pointer[num_crossings] non-zero, size = " << code_list_pointer[num_crossings]->size() << endl;
	
			for (unsigned int i = 0; i< code_list_pointer[num_crossings]->size(); i++)		
			{
				if ((*code_list_pointer[num_crossings])[i].size() != 0)
				{
					list<code_list_element>::iterator lptr = (*code_list_pointer[num_crossings])[i].begin();
					
					while (lptr != (*code_list_pointer[num_crossings])[i].end())
					{
						if (lptr->member_equivalence_class == 0)
						{
if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "find_next_root: found root with num_crossings = " << num_crossings << " index = " << i << " offset = " << distance((*code_list_pointer[num_crossings])[i].begin(),lptr) 
	      << " location = " << &*lptr << endl;
}
							return &*lptr;
						}
	
						lptr++;
					}
				}
			}
		}	
	}
	
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "find_next_root: no more root nodes, returning zero" << endl;
	
	return 0;
}


bool equivalence_class::change_to_representation(string& s)
{
	bool this_degenerate = (min_representation.find(" && ") != string::npos);
	bool s_degenerate = (s.find(" && ") != string::npos);
	
	bool change = false;
	
	if (this_degenerate)
	{
		change = false;
	}
	else if (!this_degenerate && s_degenerate)
	{
		change = true;
	}
	else if (s.length() < min_representation.length())
	{
		change = true;
	}
	else if (min_representation.length() < s.length())
	{
		change = false;
	}
	else if (min_representation.length() == 0 && s.length() == 0)
	{
		change = false;
	}
	else if (min_representation == s)  // we might have encountered the adjacent class before
	{
		change = false;
	}
	else // both codes must have the same number of crossings
	{
		vector<pair<int,int> > this_weight(2*min_num_crossings);
		vector<char> this_sign(2*min_num_crossings,'!');
				
		istringstream iss(min_representation);
		char c;		

		for (int i=0; i< 2*min_num_crossings; i++)
		{
			iss >> this_weight[i].first;
			
			iss >> c;
			if (c != ',')
				iss.putback(c);		
		}
			
		for (int i=0; i< 2*min_num_crossings; i++)
		{
			if (this_weight[i].first < 0)
				this_weight[i].second = 1;
			else
				this_weight[i].second = 0;
				
			this_weight[i].first = abs(this_weight[i].first);
		}
			
		iss >> c;
		
		while (c != '/')
			iss >> c;
			
		for (int i=0; i< min_num_crossings; i++)
		{
			iss >> c;
			while (c != '+' && c != '-')
				iss >> c;
			this_sign[i] = c;
		}
			
		
		vector<pair<int,int> > s_weight(2*min_num_crossings);
		vector<char> s_sign(2*min_num_crossings,'!');
		
		istringstream s_iss(s);
		for (int i=0; i< 2*min_num_crossings; i++)
		{
			s_iss >> s_weight[i].first;

			s_iss >> c;
			if (c != ',')
				s_iss.putback(c);
		}


		for (int i=0; i< 2*min_num_crossings; i++)
		{
			if (s_weight[i].first < 0)
				s_weight[i].second = 1;
			else
				s_weight[i].second = 0;
				
			s_weight[i].first = abs(s_weight[i].first);
		}
			
		s_iss >> c;
		
		while (c != '/')
			s_iss >> c;
			
		for (int i=0; i< min_num_crossings; i++)
		{
			s_iss >> c;
			while (c != '+' && c != '-')
				s_iss >> c;
			s_sign[i] = c;
		}
		
		
if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "change_to_representation: this min_representation " << min_representation << endl;
	debug << "change_to_representation: perm_weight ";
	for (int i=0; i< 2*min_num_crossings; i++)
		debug << '(' << this_weight[i].first << "," << (this_weight[i].second?'U':'O') << ")" << ' ';
		
	for (int i=0; i< min_num_crossings; i++)
		debug << this_sign[i] << ' ';	
	debug << endl;

	debug << "change_to_representation: s " << s << endl;
	debug << "change_to_representation: s_weight ";
	for (int i=0; i< 2*min_num_crossings; i++)
		debug << '(' << s_weight[i].first << "," << (s_weight[i].second?'U':'O') << ")" << ' ';
		
	for (int i=0; i< min_num_crossings; i++)
		debug << s_sign[i] << ' ';	
	debug << endl;

}
		
		if (this_weight < s_weight)
			change = false;
		else if (s_weight < this_weight)
			change = true;
		else
		{
			for (int i=0; i< min_num_crossings; i++)
			{
				if (this_sign[i] == '+' && s_sign[i] == '-')
				{
					change = false;
					break;
				}
				else if (this_sign[i] == '-' && s_sign[i] == '+')
				{
					change = true;
					break; 
				}
			}
		}		
	}
	
	return change;
}

void reverse_root_path(code_list_element* tree_node, code_list_element* new_parent)
{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "reverse_root_path: given parameters tree_node = " << tree_node << " new_parent = " << new_parent << endl;

	code_list_element* child = 0;
	
	while (tree_node->parent != 0)
	{

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "reverse_root_path:   tree_node " << tree_node << " current parent = " << tree_node->parent << ", parent index = " << tree_node->parent->index << ", child = " << child;
	if (child != 0)
		debug << ", index = " << child->index;
	debug << endl;
}
		code_list_element* next_tree_node = tree_node->parent;						

		if (child == 0)
			tree_node->parent = new_parent;
		else
			tree_node->parent = child; 		

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "reverse_root_path:   tree_node " << tree_node << " new parent = " << tree_node->parent;
	if (tree_node->parent != 0)
		debug << ", parent index = " << tree_node->parent->index;

	debug << ", child = " << child;		
	if (child != 0)
		debug << ", index = " << child->index;
	debug << endl;
}

		child = tree_node;
		tree_node = next_tree_node;
	}
	
	if (child == 0)
		tree_node->parent = new_parent;
	else
		tree_node->parent = child; 

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "reverse_root_path: tree_node " << tree_node << ", index = " << tree_node->index << " parent = " << tree_node->parent;
	if (tree_node->parent != 0)
		debug << ", index = " << tree_node->parent->index;
	debug << endl;
}

}

void equivalence_class_location_error(int class_identifier, string message)
{
	cout << "\nError graft_tree failed to find equivalence class including class identifier " << class_identifier << ' ';
	cout << message << endl;
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
		debug << "Error! Failed to find equivalence class including class identifier " << class_identifier << ' ';
		debug << message << endl;
}	
	exit(0);
}

/* remove_Reidemeister_II successively removes all Reidemeister II and Reidemeister I configurations from the gauss_data
   provided and returns a gauss code, written from the classical_gauss_data of the reduced gauss_orientation_data.
*/
generic_code_data remove_Reidemeister_II(gauss_orientation_data gauss_data, pair<int,int>& bigon, list_target& t)
{

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "remove_Reidemeister_II: presented with gauss_data:" ;
	print_gauss_data(gauss_data, debug, "remove_Reidemeister_II:   ");
	debug << "remove_Reidemeister_II: bigon = " << bigon.first << ' ' << bigon.second << endl;
	debug << endl;
}

	bool finished;
	generic_code_data reduced_code_data;
	
	do
	{											
		t.R2_move.push_back(bigon);
		ostringstream oss;
		
		for (int j=0; j<gauss_data.num_components; j++)
		{
			int term_count = 0;
			for (int k=gauss_data.start_of_component[j]; k<gauss_data.start_of_component[j]+gauss_data.num_terms_in_component[j]; k++)
			{
				bool keep_term = false;
				
				if (abs(gauss_data.classical_gauss_data[k]) < bigon.first)
				{
					keep_term = true;
					oss << gauss_data.classical_gauss_data[k];
					term_count++;
				}
				else if (abs(gauss_data.classical_gauss_data[k]) > bigon.first && abs(gauss_data.classical_gauss_data[k]) < bigon.second)
				{
					keep_term = true;
					if (gauss_data.classical_gauss_data[k] > 0)
						oss << gauss_data.classical_gauss_data[k]-1;									
					else
						oss << gauss_data.classical_gauss_data[k]+1;									
					term_count++;
				}
				else if (abs(gauss_data.classical_gauss_data[k]) > bigon.second)
				{
					keep_term = true;
					if (gauss_data.classical_gauss_data[k] > 0)
						oss << gauss_data.classical_gauss_data[k]-2;									
					else
						oss << gauss_data.classical_gauss_data[k]+2;									
					term_count++;
				}
				
				if (keep_term && (k < gauss_data.start_of_component[j]+gauss_data.num_terms_in_component[j]-1))
					oss << ' ';
			}
			
			if (term_count > 0 && j<gauss_data.num_components-1)
				oss << ',';
		}
							
		oss << '/';
		
		for (int j=0; j<gauss_data.num_terms/2; j++)
		{
			if (j != bigon.first-1 && j != bigon.second-1)
			{
				if (gauss_data.classical_crossing_sign[j] > 0)
					oss << "+ ";
				else									
					oss << "- ";
			}
		}
		
		string reduced_code = oss.str();
		size_t pos = reduced_code.find(",/");
		if (pos != string::npos)
			reduced_code.erase(pos,1);
														
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "Reidemeister II reduced Gauss code " << reduced_code << endl;

		read_gauss_code(reduced_code_data, reduced_code);

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "read reduced_code_data ";
	print_code_data(debug,reduced_code_data);
	debug << endl;
}

		/* At this stage, we may have disconnected the diagram by removing the bigon.  Each connected component is comprised of a set of
		   unicursal components with the property that the set of over and under terms amongst them includes all of the over and under terms for the 
		   crossings of the component and no others.  That is, the number of occurrences of each crossing amongst those unicursal components is even. 
		*/
							
		if (reduced_code_data.num_crossings == 0)
		{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "no crosings remaining setting finished true" << endl;
			finished = true;
		}
		else
		{
			finished = true;

			/* remove any Reidemeister I configurations resulting from the Reidemeister II move */
			remove_Reidemeister_I(reduced_code_data, t);	

			if (reduced_code_data.num_crossings > 0)
			{
				gauss_data = gauss_orientation_data(reduced_code_data);

				if (classical_Reidemeister_II_present(gauss_data,bigon))
				{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "remove_Reidemeister_II: detected Reidemeister II, bigon crossings " << bigon.first << " and " << bigon.second << endl;

					if (bigon.first > bigon.second)
						swap(bigon.first, bigon.second);

					finished = false;
				}
			}
		}		
	} while(!finished);
	
	return reduced_code_data;
}

string minimum_class_representative(vector<list<code_list_element> >& code_list, list<equivalence_class>& class_list, int num_crossings, bool hashed, string s)
{
	string min_representation;
	bool found = false;
	int class_identifier = -1;


	if (hashed)
	{
		size_t hash_val = hash<string>()(s);
		hash_val >>= 64-shift_power;
							
		list<code_list_element>& hash_list = code_list[hash_val];
		list<code_list_element>::iterator sptr = find_if(hash_list.begin(),hash_list.end(),code_equals(s));
	
		if (sptr != hash_list.end())
		{							
			int offset = distance(hash_list.begin(),sptr);

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "minimum_class_representative: found " << s << " at code_list index " << hash_val << ", offset " << offset << " of the list" << endl;
			class_identifier = sptr->member_equivalence_class;
			found = true;
		}
		else
		{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "Error!  minimum_class_representative failed to find " << s << " at hash_value " << hash_val << "in list for " << num_crossings << " crossings" << endl;
		}
	}
	else
	{		
		for (size_t k=0; k < code_list.size(); k++)
		{
			list<code_list_element>::iterator sptr =  code_list[k].begin();
			if (sptr->code == s)
			{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "minimum_class_representative: found " << s << " at location " << k << ", offset " << 0 << " of the list" << endl;
	
				class_identifier = sptr->member_equivalence_class;
				found = true;
				break;
			}
			else
			{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "Error!  minimum_class_representative failed to find " << s << " in list for " << num_crossings << " crossings" << endl;
			}
		}
	}
   
    if (found)
	{
			list<equivalence_class>::iterator ec_ptr = find_if(class_list.begin(),class_list.end(),class_identifier_includes(class_identifier));
			if (ec_ptr == class_list.end())
				equivalence_class_location_error(class_identifier, "looking for minimum class representative");

			min_representation = ec_ptr->min_representation;
	}
	else
    {
		cout << "Error! minimum_class_representative failed to find " << s << endl;
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "minimum_class_representative: Error! minimum_class_representative failed to find " << s << endl;
	
		exit(0);
	}

		return min_representation;
}
