/************************************************************************
                  Support functions for glist and gprocess

**************************************************************************/
using namespace std;
#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>
#include <algorithm>
#include <iomanip>

/********************* External variables ***********************/
extern ofstream     debug;



#include <matrix.h>
#include <debug-control.h>
#include <generic-code.h>
#include <gauss-orientation.h>



string gauss_plane_reflect (string c, bool unoriented)
{
if (debug_control::DEBUG >= debug_control::BASIC)
	debug << "gauss_plane_reflect: presented with code: " << c << endl;

	generic_code_data code_data;
	read_gauss_code(code_data,c);
	
	for (int i=0; i< code_data.num_crossings; i++)
	{
		if (code_data.code_table[LABEL][i] == generic_code_data::POSITIVE)
			code_data.code_table[LABEL][i] = generic_code_data::NEGATIVE;
		else
			code_data.code_table[LABEL][i] = generic_code_data::POSITIVE;
	}
	
if (debug_control::DEBUG >= debug_control::BASIC)
{
	debug << "gauss_plane_reflect: reflected code: ";
	write_gauss_code(debug,code_data);
	debug << endl;
}
	string preferred_reflected_code = over_preferred_gauss_code(code_data, unoriented);

if (debug_control::DEBUG >= debug_control::BASIC)
	debug << "gauss_plane_reflect: preferred_reflected_code: " << preferred_reflected_code << endl;
	
	return preferred_reflected_code;
}

string gauss_inverse (string c, vector<bool>& r_flag, bool unoriented)
{
	
if (debug_control::DEBUG >= debug_control::BASIC)
	debug << "gauss_inverse: presented with code: " << c << endl;

	generic_code_data code_data;
	read_gauss_code(code_data,c);
	gauss_orientation_data gdata(code_data);
	ostringstream oss;
	
	/* invert those components indicated by r_flag */
	for (int i=0; i<gdata.num_components; i++)
	{
		if (r_flag[i])
		{
			for (int j=gdata.num_terms_in_component[i]-1; j>=0; j--)
				oss << gdata.classical_gauss_data[gdata.start_of_component[i]+j] << ' ';
		}
		else
		{
			for (int j=0; j< gdata.num_terms_in_component[i]; j++)
				oss << gdata.classical_gauss_data[gdata.start_of_component[i]+j] << ' ';
		}
		
		if (i < gdata.num_components-1)
			oss << ',';
	}
	oss << c.substr(c.find('/'));
	
if (debug_control::DEBUG >= debug_control::BASIC)
	debug << "gauss_inverse: inverse code: " << oss.str() << endl;
	
	generic_code_data inverse_code_data;
	read_gauss_code(inverse_code_data,oss.str());		
	string preferred_inverse_code = over_preferred_gauss_code(inverse_code_data, unoriented);

if (debug_control::DEBUG >= debug_control::BASIC)
	debug << "gauss_inverse: preferred_inverse_code: " << preferred_inverse_code << endl;

	return preferred_inverse_code;
}

string gauss_line_reflect (string c, bool unoriented)
{
if (debug_control::DEBUG >= debug_control::BASIC)
	debug << "gauss_line_reflect: presented with code: " << c << endl;

	string reflected_code = c;
	bool signs = false;
	
	for (unsigned int i=0; i< c.length(); i++)
	{
		if (c[i] == '/')
		{
			signs = true;
			continue;
		}
			
		if(signs)
		{
			if (c[i] == '+')
				reflected_code[i] = '-';
			else if (c[i] == '-')
				reflected_code[i] = '+';
		}
	}

	generic_code_data reflected_code_data;
	read_gauss_code(reflected_code_data,reflected_code);		
	string preferred_reflected_code = over_preferred_gauss_code(reflected_code_data, unoriented);
	
if (debug_control::DEBUG >= debug_control::BASIC)
	debug << "gauss_line_reflect: preferred_reflected_code: " << preferred_reflected_code << endl;

	return preferred_reflected_code;
}


